#!/usr/bin/python

#-------------------------------------------------------------------------------
#
#   Copyright (C) 2017 Cumulus Networks, Inc. All rights reserved
#
#   This software is subject to the Cumulus Networks End User License
#   Agreement available  at the following locations:
#
#   Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#   Cumulus Linux systems: /usr/share/cumulus/EULA.txt
#
#-------------------------------------------------------------------------------

"""
This module implements the Accton chassis daemon, which provides a loop-free
management network within the chassis and handles thermal managment.
"""


#-------------------------------------------------------------------------------
#
#   Import the necessary modules
#
#-------------------------------------------------------------------------------

try:
    import argparse
    import chassisd.chassisd_bcm53xx as bcm
    import ctypes
    import dpkt
    import errno
    import fcntl
    import glob
    import logging
    import os
    import select
    import socket
    import struct
    import subprocess
    import sys
    import traceback
    import threading
    import time
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")


#-------------------------------------------------------------------------------
#
#   Global Variables
#
#-------------------------------------------------------------------------------

# The handle used for logging messages
logger = logging.getLogger("cumulus-chassisd")

# Largest slot ID value
_MAX_SLOT_ID = 48

# Parser defaults
_DEFAULT_FC_MCAST           = "01:10:b5:00:00:01"
_DEFAULT_LC_MCAST           = "01:10:b5:00:00:02"
_DEFAULT_STATIC_TOPO        = False
_DEFAULT_BACKPLANE          = 1
_DEFAULT_IFNAME             = "eth0"
_DEFAULT_VLAN               = 4088
_DEFAULT_MSG_AGE            = 1
_DEFAULT_MAX_AGE            = 15
_DEFAULT_FWD_DELAY          = 3
_DEFAULT_HELLO_TIME         = 200
_DEFAULT_THERMAL_FREQ       = 15
_DEFAULT_NO_SHUTDOWN        = False

# The maximum and minimum possible thermal temperatures (in degrees C)
_MAX_TEMP                   = 127
_MIN_TEMP                   = -128

# from include/uapi/linux/if_ether.h
_ETH_P_ALL = 3

# from include/linux/socket.h
_SO_ATTACH_FILTER = 26


#-------------------------------------------------------------------------------
#
#   Chassis Topology Protocol classes (data structures)
#
#-------------------------------------------------------------------------------

# A bridge identifier is a priority and mac address. The priority is 4096 plus
# the slot number. The mac address is the mac of the management interface.
class brId:
    def __init__(self, prio=0, addr="00:00:00:00:00:00"):
        self.prio = prio
        self.addr = addr

# Bridge information is the slot id of the bridge, an age of the info and a
# bridge id of that bridge.
class brInfo:
    def __init__(self, slotId=None, age=0, id=brId()):
        self.slotId = slotId
        self.age    = age
        self.id     = id

# This class defines the information we keep for the CTP protocol
class ctpInfo:

    # adminState
    _CTP_ENABLED           = 0
    _CTP_DISABLED          = 1

    # state
    _STATE_BLOCK           = 0
    _STATE_LISTENING       = 1
    _STATE_LEARNING        = 2
    _STATE_FORWARDING      = 3
    _STATE_DISABLE         = 255

    # CPDU types
    _CPDU_TYPE_CONFIG      = 0x00
    _CPDU_TYPE_HELLO       = 0xFF
    _CPDU_TYPE_REPLY       = 0x01
    _CPDU_TYPE_LC_CHANGE   = 0x03
    _CPDU_TYPE_FC_CHANGE   = 0x04
    _CPDU_TYPE_TCN         = 0x80
    _CPDU_TYPE_RSTP        = 0x02

    # CPDU flags
    _CPDU_FLAGS_TCN        = 0x01
    _CPDU_FLAGS_PROPOSAL   = 0x02
    _CPDU_FLAGS_PORT_ROLE  = 0x0C
    _CPDU_FLAGS_ROOT_FC    = 0x0C
    _CPDU_FLAGS_DESG_LC    = 0x08
    _CPDU_FLAGS_OTHER_ROLE = 0x04
    _CPDU_FLAGS_LEARN      = 0x10
    _CPDU_FLAGS_FORWARD    = 0x20
    _CPDU_FLAGS_AGREEMENT  = 0x40
    _CPDU_FLAGS_TCN_ACK    = 0x80

    def __init__(self, adminState=_CTP_DISABLED, state=_STATE_DISABLE, brId=brId(),
                 brMsgAge=_DEFAULT_MSG_AGE, brMaxAge=_DEFAULT_MAX_AGE,
                 brFwdDelay=_DEFAULT_FWD_DELAY, brHelloTime=_DEFAULT_HELLO_TIME,
                 rootPortId=0, rootLCPortId=0, rootBrId=brId(), rootPathCost=0,
                 designBrId=brId(), designPort=0, brPort=0, rootMsgAge=_DEFAULT_MSG_AGE,
                 rootMaxAge=_DEFAULT_MAX_AGE, rootFwdDelay=_DEFAULT_FWD_DELAY,
                 rootHelloTime=_DEFAULT_HELLO_TIME, timeSinceTopoChange=0,
                 topoChangeCount=0, topoChange=False,
                 brExist=[brInfo(i+1) for i in xrange(_MAX_SLOT_ID)]):

        self.adminState             = adminState
        self.state                  = state
        self.brId                   = brId
        self.brMsgAge               = brMsgAge
        self.brMaxAge               = brMaxAge
        self.brFwdDelay             = brFwdDelay
        self.brHelloTime            = brHelloTime
        self.rootPortId             = rootPortId
        self.rootLCPortId           = rootLCPortId
        self.rootBrId               = rootBrId
        self.rootPathCost           = rootPathCost
        self.designBrId             = designBrId
        self.designPort             = designPort
        self.brPort                 = brPort
        self.rootMsgAge             = rootMsgAge
        self.rootMaxAge             = rootMaxAge
        self.rootFwdDelay           = rootFwdDelay
        self.rootHelloTime          = rootHelloTime
        self.timeSinceTopoChange    = timeSinceTopoChange
        self.topoChangeCount        = topoChangeCount
        self.topoChange             = topoChange
        self.brExist                = brExist

class tempSensorInfo:
    def __init__(self, warn=_MAX_TEMP, shutdown=_MAX_TEMP, temp=_MIN_TEMP, errors=0):
        self.warn       = warn
        self.shutdown   = shutdown
        self.temp       = temp
        self.errors     = errors

class fanCtrlInfo:
    def __init__(self, idx=0, asic=tempSensorInfo(), cpu=tempSensorInfo(), lm75a=tempSensorInfo(),
                 lm75b=tempSensorInfo(), lm75c=tempSensorInfo(), lm75d=tempSensorInfo()):
        self.idx    = idx
        self.asic   = asic
        self.cpu    = cpu
        self.lm75a  = lm75a
        self.lm75b  = lm75b
        self.lm75c  = lm75c
        self.lm75d  = lm75d

# This class holds the temperature information which a fabric card (the local
# master) is responsible for monitoring.
class lmTempInfo:
    def  __init__(self,
                  fca=[fanCtrlInfo(idx=0), fanCtrlInfo(idx=1)],
                  fcb=[fanCtrlInfo(idx=0), fanCtrlInfo(idx=1)],
                  lca=[fanCtrlInfo(idx=0), fanCtrlInfo(idx=1)],
                  lcb=[fanCtrlInfo(idx=0), fanCtrlInfo(idx=1)],
                  lcc=[fanCtrlInfo(idx=0), fanCtrlInfo(idx=1)],
                  lcd=[fanCtrlInfo(idx=0), fanCtrlInfo(idx=1)]):
        self.fca = fca
        self.fcb = fcb
        self.lca = lca
        self.lcb = lcb
        self.lcc = lcc
        self.lcd = lcd


#-------------------------------------------------------------------------------
#
#   dpkt classes for building/parsing packet headers/data
#
#-------------------------------------------------------------------------------

class LLC(dpkt.dpkt.Packet):
    '''
    The 802.2 LLC header
    '''
    __hdr__ = (
        ('dsap', 'B', 0),
        ('ssap', 'B', 0),
        ('ctrl', 'B', 0)
    )

class CtpData(dpkt.dpkt.Packet):
    '''
    The CTP data which follows the "STP" header
    '''
    __hdr__ = (('ver1Len', 'H', 0),)
    for i in xrange(_MAX_SLOT_ID):
        __hdr__ += (('age%d' % i, 'B', 0),)
        __hdr__ += (('pad%d' % i, 'B', 0),)
        __hdr__ += (('prio%d' % i, 'H', 0),)
        __hdr__ += (('addr%d' % i, '6s', '\x00' * 6),)


#-------------------------------------------------------------------------------
#
#   BPF socket filter and program classes (data structures)
#
#-------------------------------------------------------------------------------

# from include/uapi/linux/filter.h
class struct_sock_filter(ctypes.Structure):
    _fields_ = [('code', ctypes.c_ushort),
                ('jt', ctypes.c_ubyte),
                ('jf', ctypes.c_ubyte),
                ('k', ctypes.c_uint)]

# This goes as arg with SO_ATTACH_FILTER
class struct_sock_fprog(ctypes.Structure):
    _fields_ = [('len', ctypes.c_ushort),
                ('filter', ctypes.POINTER(struct_sock_filter))]



#-------------------------------------------------------------------------------
#
#   The chassisd_accton class
#
#-------------------------------------------------------------------------------

class chassisd_accton:
    """
    This class implements the functionality of Accton's chassisd daemon, but
    in python.
    """

    #
    # Constants
    #
    _CPLD1_SYSFS   = "/sys/bus/i2c/devices/12-0060/"
    _BCM53XX_SYSFS = "/sys/bus/platform/devices/bcm53xx.0/"

    #
    #   Code
    #
    def __init__(self, chassis, cmdline, stopEvent):
        # Save some command line parameters which we will use later
        self.chassis        = chassis
        self.cmdargs, self.cmdline = self.ParseCmdLine(cmdline)
        self.stopEvent      = stopEvent
        self.bcmPorts       = {}
        self.ctpRxT         = None
        self.ctpTxT         = None
        self.thermalT       = None
        self.slotId         = self.chassis.GetSlotNum()
        self.ctpData        = None
        self.localMasters   = [0] * 4
        self.selfPipeR      = None
        self.selfPipeW      = None
        self.sock           = None
        self.tempData       = None
        self.rxCpduInterval = {}
        self.SetBcmPorts()
        self.UpMgmtIfs()
        if self.isCpuA():
            self.SetDefaultTopology()
            self.SelectMasterFC()
            if not self.cmdargs.chasStaticTopo:
                self.GenBpfFilter()
                self.selfPipeR, self.selfPipeW = self.InitSelfPipe()
                if self.selfPipeR is None or self.selfPipeW is None:
                    logger.error("Could not create the self pipe.")
                    self.stopEvent.set()
                    return

                # Create the CTP Receive and Transmit Threads
                self.ctpRxT = threading.Thread(None, self.CtpRxT, 'CtpRxT')
                self.ctpRxT.daemon = True
                self.ctpTxT = threading.Thread(None, self.CtpTxT, 'CtpTxT')
                self.ctpTxT.daemon = True

                # Start the threads
                self.ctpRxT.start()
                self.ctpTxT.start()

        # Start the thermal plan thread
        self.thermalT = threading.Thread(None, self.ThermalT, 'ThermalT')
        self.thermalT.daemon = True
        self.thermalT.start()


    def __del__(self):
        self.ClearDefaultTopology()

    def ParseCmdLine(self, cmdline):
        parser = argparse.ArgumentParser(description="The Accton Chassid Daemon")
        parser.add_argument("--chasFcMcast", default=_DEFAULT_FC_MCAST,
                            metavar="MACADDR", help="The multicast address fabric cards send.")
        parser.add_argument("--chasLcMcast", default=_DEFAULT_LC_MCAST,
                            metavar="MACADDR", help="The multicast address line cards send.")
        parser.add_argument("--chasStaticTopo", action="store_true", default=_DEFAULT_STATIC_TOPO,
                            help="Use a static mgmt network topology")
        parser.add_argument("--chasBackplane", type=self.intBase, default=_DEFAULT_BACKPLANE,
                            metavar="BACKPLANE_VERSION", help="The version of the chassis backplane.")
        parser.add_argument("--chasIfName", default=_DEFAULT_IFNAME,
                            metavar="IFNAME", help="The interface upon which to run CTP.")
        parser.add_argument("--chasVlan", type=self.intBase, default=_DEFAULT_VLAN,
                            metavar="VLAN_ID", help="The management VLAN id.")
        parser.add_argument("--chasMsgAge", default=_DEFAULT_MSG_AGE,
                            metavar="", help="The message age.")
        parser.add_argument("--chasMaxAge", default=_DEFAULT_MAX_AGE,
                            metavar="", help="The maximum age.")
        parser.add_argument("--chasFwdDelay", default=_DEFAULT_FWD_DELAY,
                            metavar="", help="The forwarding delay.")
        parser.add_argument("--chasHelloTime", default=_DEFAULT_HELLO_TIME,
                            metavar="MILLISECS", help="The hello time.")
        parser.add_argument("--chasThermalFreq", default=_DEFAULT_THERMAL_FREQ,
                            metavar="SECONDS", help="The frequency of the thermal manager.")
        parser.add_argument("--chasNoShutdown", action="store_true", default=_DEFAULT_NO_SHUTDOWN,
                            help="Don't shutdown cards which are over shutdown threshold.")
        args, cmdline = parser.parse_known_args(cmdline)
        return(args, cmdline)

    def intBase(self, intStr):
        '''
        Allow an integer to be specified with standard base representations:
        37 = decimal, 0x25 = hex, 045 = octal, 0b100101 = binary
        '''
        try:
            intVal = int(intStr,0)
        except:
            msg = "%s is not a valid integer" % (intStr,)
            raise argparse.ArgumentTypeError(msg)
        return intVal

    def SignalThreads(self):
        if self.selfPipeW is not None:
            try:
                os.write(self.selfPipeW, 'x')
            except os.error:
                pass

    def JoinThreads(self):
        if self.ctpRxT is not None and self.ctpRxT.is_alive():
            self.ctpRxT.join()
        if self.ctpTxT is not None and self.ctpTxT.is_alive():
            self.ctpTxT.join()
        if self.thermalT is not None and self.thermalT.is_alive():
            self.thermalT.join()

    def isCpuA(self):
        """
        Returns true if running on CPU A, and false if not
        """
        return self.slotId % 2 == 1

    def isLowerChassis(self):
        """
        Returns true if running on the lower chassis
        """
        slot = self.slotId
        return 17 <= slot <= 32 or 41 <= slot <= 48

    def isFirstLcOrFc(self):
        """
        Returns true if running on CPUA of the first fabric card or line card in
        the uppoer or lower chassis.
        """
        return self.slotId in [1, 17, 33, 41]

    def isFabricCard(self, slot=None):
        """
        Returns true if the provided slot number is a fabric card.
        """
        if slot is None:
            slot = self.slotId
        return slot > 32

    def WriteSysfsNode(self, node, value):
        """
        Writes a sysfs node
        """
        logger.debug("Chassisd - Writing sysfs node %s with value %s" % (node, str(value)))
        try:
            with open(node, 'w') as f:
                f.write(str(value))
        except IOError as e:
            logger.error("Chassisd - I/O error writing %s to %s: %s" %
                         (value, node, e))

    def ReadSysfsNode(self, node):
        """
        Reads a sysfs node
        """
        value = None
        try:
            with open(node) as f:
                value = f.readline().strip()
        except (AttributeError, IOError) as e:
            value = None
        logger.debug("Chassisd - Read %s from sysfs node %s" % (str(value), node))
        return value

    def WriteBcm53xx(self, node, value):
        """
        Write to the Bcm53xx device through the sysfs
        """
        self.WriteSysfsNode(chassisd_accton._BCM53XX_SYSFS + node, value)

    def WriteCpld1(self, node, value):
        """
        Write to the CPLD1 device through the sysfs
        """
        self.WriteSysfsNode(chassisd_accton._CPLD1_SYSFS + node, value)

    def InitCtpData(self, addr):
        """
        Initialize the ctpData structure.
        """
        # Initialize the CTP information
        adminState = ctpInfo._CTP_DISABLED
        state      = ctpInfo._STATE_DISABLE
        if not self.cmdargs.chasStaticTopo and self.isCpuA():
            adminState = ctpInfo._CTP_ENABLED
            state      = ctpInfo._STATE_BLOCK

        brPrio             = 4096 + self.slotId
        rootPortId         = 0
        rootLCPortId       = self.slotId
        rootBrId           = brId()
        rootPathCost       = 2000
        designBrId         = brId(brPrio, addr)
        designPort         = self.slotId
        brPort             = self.slotId
        if self.isFabricCard():
            rootPortId         = self.slotId
            rootLCPortId       = 0
            rootBrId           = brId(brPrio, addr)
            rootPathCost       = 2000
            designBrId         = brId()
            designPort         = 1

        cd = ctpInfo(adminState=adminState, state=state, brId=brId(brPrio, addr),
                     brMsgAge=self.cmdargs.chasMsgAge, brMaxAge=self.cmdargs.chasMaxAge,
                     brFwdDelay=self.cmdargs.chasFwdDelay, brHelloTime=self.cmdargs.chasHelloTime,
                     rootPortId=rootPortId, rootLCPortId=rootLCPortId, rootBrId=rootBrId,
                     rootPathCost=rootPathCost, designBrId=designBrId, designPort=designPort,
                     brPort=brPort, rootMsgAge=self.cmdargs.chasMsgAge, rootMaxAge=self.cmdargs.chasMaxAge,
                     rootFwdDelay=self.cmdargs.chasFwdDelay, rootHelloTime=self.cmdargs.chasHelloTime)

        cd.brExist[self.slotId-1].age  = 5*self.cmdargs.chasFwdDelay + 1
        cd.brExist[self.slotId-1].id   = brId(brPrio, addr)

        self.ctpData = cd
        if not self.cmdargs.chasStaticTopo:
            self.CardEnterBlock()
        if self.isFabricCard:
            self.SelectMasterFC()

    def InitTempData(self):
        """
        Initialize the tempData structure.
        """
        temps     = { "cpu" : 0x3f, "asic" : 0x51, "lm75a" : 0x3b, "lm75b" : 0x40, "lm75c" : 0x44, "lm75d" : 0x3d }
        warnTemps = { "cpu" : 0x40, "asic" : 0x52, "lm75a" : 0x3c, "lm75b" : 0x41, "lm75c" : 0x45, "lm75d" : 0x3e }
        shutTemps = { "cpu" : 0x5f, "asic" : 0x69, "lm75a" : 0x55, "lm75b" : 0x5a, "lm75c" : 0x5e, "lm75d" : 0x57 }
        fcd = {}
        for card in ["fca", "fcb", "lca", "lcb", "lcc", "lcd"]:
            fci=[]
            for fctrl in [0, 1]:
                tsd = {}
                for sensor in ["asic", "cpu", "lm75a", "lm75b", "lm75c", "lm75d"]:
                    temp     = temps.get(sensor, _MAX_TEMP)
                    warnTemp = warnTemps.get(sensor, _MAX_TEMP)
                    shutTemp = shutTemps.get(sensor, _MAX_TEMP)
                    tsd[sensor] = tempSensorInfo(warn=warnTemp, shutdown=shutTemp, temp=temp)
                fci.append(fanCtrlInfo(idx=fctrl, asic=tsd["asic"], cpu=tsd["cpu"],
                                       lm75a=tsd["lm75a"], lm75b=tsd["lm75b"], 
                                       lm75c=tsd["lm75c"], lm75d=tsd["lm75d"]))
            fcd[card] = fci
        self.tempData = lmTempInfo(fca=fcd["fca"], fcb=fcd["fcb"],
                                   lca=fcd["lca"], lcb=fcd["lcb"], 
                                   lcc=fcd["lcc"], lcd=fcd["lcd"], )

    def SetBcmPorts(self):
        """
        Fill out the bcmPorts dictionary for the type of card/system we are
        running upon.
        """
        if self.isFabricCard():
            # First two fabric cards on upper or lower chassis?
            if self.slotId / 4 % 2 == 0:
                if self.cmdargs.chasBackplane == 1:
                    self.bcmPorts = { "CPUA" : 14, "CPUB" : 15, "FC"   : 11, "RJ45" : 9,
                                      "LC0"  : 5,  "LC1"  : 4,  "LC2"  : 7,  "LC3"  : 6,
                                      "LC4"  : 1,  "LC5"  : 0,  "LC6"  : 3,  "LC7"  : 2,
                                      "NON1" : 8,  "NON2" : 10, "NON3" : 12, "NON4" : 13 }
                else:
                    self.bcmPorts = { "CPUA" : 14, "CPUB" : 15, "FC"   : 11, "RJ45" : 9,
                                      "LC0"  : 4,  "LC1"  : 5,  "LC2"  : 6,  "LC3"  : 7,
                                      "LC4"  : 0,  "LC5"  : 1,  "LC6"  : 2,  "LC7"  : 3,
                                      "NON1" : 8,  "NON2" : 10, "NON3" : 12, "NON4" : 13 }
            else:
                if self.cmdargs.chasBackplane == 1:
                    self.bcmPorts = { "CPUA" : 14, "CPUB" : 15, "FC"   : 11, "RJ45" : 9,
                                      "LC0"  : 1,  "LC1"  : 0,  "LC2"  : 3,  "LC3"  : 2,
                                      "LC4"  : 5,  "LC5"  : 4,  "LC6"  : 7,  "LC7"  : 6,
                                      "NON1" : 8,  "NON2" : 10, "NON3" : 12, "NON4" : 13 }
                else:
                    self.bcmPorts = { "CPUA" : 14, "CPUB" : 15, "FC"   : 11, "RJ45" : 9,
                                      "LC0"  : 0,  "LC1"  : 1,  "LC2"  : 2,  "LC3"  : 3,
                                      "LC4"  : 4,  "LC5"  : 5,  "LC6"  : 6,  "LC7"  : 7,
                                      "NON1" : 8,  "NON2" : 10, "NON3" : 12, "NON4" : 13 }
        else:
            self.bcmPorts = { "CPUA" : 0,  "CPUB" : 7,
                              "FC0"  : 2,  "FC1"  : 3,  "FC2"  : 4,  "FC3"  : 5,
                              "NON1" : 1,  "NON2" : 6 }

    def ResetBcm53xx(self):
        """
        Reset the Bcm53xx chip via the reset line on the CPLD
        """
        self.WriteCpld1("reset_bcm53xx", 1)
        time.sleep(0.020)
        self.WriteCpld1("reset_bcm53xx", 0)

    def SetVlanTableEntry(self, vlanId=0, valid=0, spt=0, fwdMap=0, untagMap=0):
        """
        Writes a VLAN table entry in the BCM53xx VLAN table
        """
        self.WriteBcm53xx("vlanTblAddr", vlanId)
        entry = untagMap << 23 | fwdMap << 6 | spt << 1 | valid
        self.WriteBcm53xx("vlanTblEntry", entry)
        self.WriteBcm53xx("vlanTblCntrl", 0x80)

    def MakeMgmtVlanInternal(self):
        """
        Configures the BCM53xx so that the management VLAN does not go out the
        RJ45 port.
        """
        fwdMap = ((1 << 17) - 1) & ~(1 << self.bcmPorts["RJ45"])
        self.SetVlanTableEntry(vlanId=self.cmdargs.chasVlan, valid=1, fwdMap=fwdMap)
        self.WriteBcm53xx("dot1QGlbCntrl",
                          bcm._DOT1Q_GLB_CNTRL_EN        |\
                          bcm._DOT1Q_GLB_CNTRL_LEARN_IVL |\
                          bcm._DOT1Q_GLB_CNTRL_1P_CHANGE_VID)

    def SetInitialPhy(self):
        """
        Set the initial state of the PHY. First set the default value in the
        BMC53xx external PHY scan control register, then reset the phy, and
        finally set the autonegotation advertisement.
        """
        self.WriteBcm53xx("extPhyCntrl", bcm._EXT_PHY_CTRL_ADDR_INC)
        self.WriteCpld1("reset_mgmt_phy", 1)
        time.sleep(0.001)
        self.WriteCpld1("reset_mgmt_phy", 0)
        self.WriteBcm53xx("port10AutoNegAdv",
                           bcm._PORT_AUTO_NEG_ADVER_100BASET4 |\
                           bcm._PORT_AUTO_NEG_ADVER_100FD     |\
                           bcm._PORT_AUTO_NEG_ADVER_100HD     |\
                           bcm._PORT_AUTO_NEG_ADVER_10FD      |\
                           bcm._PORT_AUTO_NEG_ADVER_10HD      |\
                           bcm._PORT_AUTO_NEG_ADVER_SEL_8023)

    def SetSGMIIMode(self, mode):
        """
        Configure the SGMII interfaces on a card
        """
        for port in self.bcmPorts:
            if port.startswith("LC") or port.startswith("FC"):
                self.WriteBcm53xx("port%dSGMIICntrl" % (self.bcmPorts[port],), mode)

    def SetStaticARL(self):
        """
        Configure the multicast address used for card-card communication
        A fabric card receives the line card multicast address, except on the
        lower chassis, where this is reversed.
        """
        mcAddr = self.cmdargs.chasLcMcast if self.isFabricCard() else self.cmdargs.chasFcMcast
        if self.isLowerChassis():
            mcAddr = self.cmdargs.chasFcMcast if self.isFabricCard() else self.cmdargs.chasLcMcast
        mcVect = 1 << self.bcmPorts["CPUA"]

        self.WriteBcm53xx("globalArlCfg", bcm._GLOBAL_ARL_CFG_MPORT_ADDR_EN)
        self.WriteBcm53xx("bpduMcastAddr", mcAddr)
        self.WriteBcm53xx("multiportAddr1", mcAddr)
        self.WriteBcm53xx("multiportVec1", mcVect)

    def SetReservedMulticastFwd(self):
        """
        Configure which of the reserved STP multicast address ranges are forwarded
        and which are dropped.
        """
        node = "resMcastCntrl96" if self.isFabricCard() else "resMcastCntrl89"
        drop13 = bcm._RES_MCAST_CNTRL_MUL3_DROP | bcm._RES_MCAST_CNTRL_MUL1_DROP
        drop14 = bcm._RES_MCAST_CNTRL_MUL4_DROP | bcm._RES_MCAST_CNTRL_MUL1_DROP
        stpFwd = drop14 if self.isFabricCard() else drop13
        if self.isLowerChassis():
            stpFwd = drop13 if self.isFabricCard() else drop14
        self.WriteBcm53xx(node, stpFwd)

    def InitialPortStpState(self):
        """
        Configure the initial port STP state. All ports are set to forwarding
        except:
           1. Unused ports, which are disabled
           2. On a line card, if the static topology is enabled, the ports on
              line cards other then LC0 that are connect to fabric cards other
              than FC0 are blocked.
           3. On a line card, in dynamic topology mode, all ports to the fabric
              cards are blocked.
        """
        for port in self.bcmPorts:
            state = bcm._PORT_CTRL_STP_STATE_FORWARDING
            if port.startswith("NON"):
                state = bcm._PORT_CTRL_STP_STATE_DISABLED
            if not self.isFabricCard():
                if self.cmdargs.chasStaticTopo:
                    if not self.isFirstLcOrFc() and port in ["FC1", "FC2", "FC3"]:
                        state = bcm._PORT_CTRL_STP_STATE_BLOCKING
                else:
                    if port.startswith("FC"):
                        state = bcm._PORT_CTRL_STP_STATE_BLOCKING
            self.ConfigPortStp(self.bcmPorts[port], state)

    def ConfigPortStp(self, port, state):
        """
        Configure the STP state of a port
        """
        self.WriteBcm53xx("port%dCntrl" % (port), state)

    def ConfigAllPortsStp(self, state):
        """
        Set all of the ports to a line card or fabric card to the specified state
        """
        for port in self.bcmPorts:
            if port.startswith("FC") or port.startswith("LC"):
                self.ConfigPortStp(self.bcmPorts[port], state)

    def DumpBrExist(self):
        logger.debug("brExist:")
        for card in self.ctpData.brExist:
            logger.debug("   %d : age %d, prio %d, addr %s" % (card.slotId, card.age, card.id.prio, card.id.addr))

    def SelectNewRootLC(self, oldRootLC):
        """
        Look through our brExist structure for the current best line card. Return
        True if one is found, otherwise False.
        """
        #self.DumpBrExist()
        be = self.ctpData.brExist
        #be[oldRootLC-1].age = 0
        for card in be:
            if card.slotId != self.slotId and card.age >= 3:
                self.ctpData.rootLCPortId  = card.slotId
                self.ctpData.designPort    = card.slotId
                self.ctpData.designBrId    = brId(card.id.prio, card.id.addr)
                logger.debug("Found new root line card: %d" % (card.slotId,))
                return True
        return False

    def UpdateLCRoot(self, ctp):
        """
        Update this card's notion of the root and designated cards
        """
        cd = self.ctpData
        rootBrId     = self.StrToBrId(ctp.root_id)
        rootSlotId   = rootBrId.prio & 0x00FF
        designBrId   = self.StrToBrId(ctp.bridge_id)
        designSlotId = designBrId.prio & 0x00FF
        if len(ctp.data) < CtpData.__hdr_len__:
            return
        cpduData = CtpData(ctp.data)
        if (rootSlotId > 0) and cpduData["age%d" % (rootSlotId-1,)] > 0:
            if rootSlotId < cd.rootPortId or cd.rootPortId == 0:
                cd.rootPortId   = rootSlotId
                cd.rootPathCost = ctp.root_path
                cd.rootBrId     = rootBrId
        if (designSlotId > 0) and cpduData["age%d" % (designSlotId-1,)] > 0:
            if ctp.type == ctpInfo._CPDU_TYPE_HELLO or \
               designSlotId < cd.rootLCPortId or cd.rootLCPortId == 0:
                cd.rootLCPortId = designSlotId
                cd.designPort = designSlotId
                cd.designBrId = designBrId

    def UpdateFCDesignated(self, ctp):
        """
        Update this card's notion of the root and designated cards
        """
        cd = self.ctpData
        rootBrId     = self.StrToBrId(ctp.root_id)
        rootSlotId   = rootBrId.prio & 0x00FF
        designBrId   = self.StrToBrId(ctp.bridge_id)
        designSlotId = designBrId.prio & 0x00FF
        if len(ctp.data) < CtpData.__hdr_len__:
            return
        cpduData = CtpData(ctp.data)
        forwarding = ctp.flags & ctpInfo._CPDU_FLAGS_FORWARD

        if (rootSlotId > 0) and cpduData["age%d" % (rootSlotId-1,)] > 0:
            if forwarding or rootSlotId < cd.rootPortId or cd.rootPortId == 0:
                cd.rootPortId   = rootSlotId
                cd.rootPathCost = ctp.root_path
                cd.rootBrId     = rootBrId
        if (designSlotId > 0) and cpduData["age%d" % (designSlotId-1,)] > 0:
            if designSlotId < cd.rootLCPortId or cd.rootLCPortId == 0:
                cd.rootLCPortId = designSlotId
                cd.designPort = designSlotId
                cd.designBrId = designBrId

    def CheckCpduNewIds(self, ctp):
        """
        Check if the CPDU contains differnet Root or Designate ID than we have.
        """
        cd = self.ctpData
        rootSlotId   = self.StrToBrId(ctp.root_id).prio & 0x00FF
        designSlotId = self.StrToBrId(ctp.bridge_id).prio & 0x00FF
        return (rootSlotId != cd.rootBrId.prio & 0x00FF) or (designSlotId != cd.designBrId.prio & 0x00FF)

    def LCSyncToHello(self, ctp):
        """
        Sync our line card state to the hello packet
        """
        cd = self.ctpData
        if len(ctp.data) < CtpData.__hdr_len__:
            return
        cpduData = CtpData(ctp.data)
        dBridgeSlotId     = self.StrToBrId(ctp.bridge_id).prio & 0x00FF
        origDBridgeSlotId = cd.designBrId.prio & 0x00FF
        myDBridgeSlotId   = cd.brId.prio & 0x00FF
        origRootLC        = cd.rootLCPortId

        if dBridgeSlotId > 0 and dBridgeSlotId != origDBridgeSlotId and cpduData["age%d" % (dBridgeSlotId-1,)] > 0:
            cd.rootLCPortId = dBridgeSlotId
            cd.designPort   = dBridgeSlotId
            cd.designBrId   = self.StrToBrId(ctp.bridge_id)
            if origRootLC == origDBridgeSlotId or myDBridgeSlotId == dBridgeSlotId:
                self.CardEnterForwarding()

    def ClearDesignedLC(self):
        """
        Remove information about the designated LC for this fabric card.
        """
        cd = self.ctpData
        cd.rootLCPortId  = 0
        cd.rootPortId    = self.slotId
        cd.rootBrId      = brId(cd.brId.prio, cd.brId.addr)
        cd.rootPathCost  = 2000
        cd.designBrId    = brId()
        cd.designPort    = 1
        cd.brPort        = self.slotId

    def ClearRootFC(self):
        """
        Remove information about the root fabric card from this line card.
        """
        cd = self.ctpData
        cd.rootLCPortId  = self.slotId
        cd.rootPortId    = 0
        cd.rootBrId      = brId()
        cd.rootPathCost  = 2000
        cd.designBrId    = brId(cd.brId.prio, cd.brId.addr)
        cd.designPort    = self.slotId
        cd.brPort        = self.slotId

    def ClearBrExists(self):
        """
        Remove information about all other cards
        """
        cd = self.ctpData
        cd.brExist = [brInfo(i+1) for i in xrange(_MAX_SLOT_ID)]
        cd.brExist[self.slotId-1].age  = 5*self.cmdargs.chasFwdDelay + 1
        cd.brExist[self.slotId-1].id   = brId(cd.brId.prio, cd.brId.addr)

    def DecBrExists(self):
        """
        Decement the age of all of the bridges we know about, except us.
        """
        for br in self.ctpData.brExist:
            if br.slotId == self.slotId:
                br.age = 5*self.cmdargs.chasFwdDelay + 1
            elif br.age > 0:
                br.age -= 1

    def UpdateBrExists(self, cpduData):
        """
        Update our bridge neighbor information based on the received data
        """
        cd = self.ctpData
        for br in cd.brExist:
            if br.slotId != self.slotId:
                cpduAge = cpduData["age%d" % (br.slotId-1,)]
                if cpduAge > 0 and br.age < cpduAge:
                    br.age = cpduAge
                    macStr = self.MacByteStrToStr(cpduData["addr%d" % (br.slotId-1,)])
                    br.id = brId(cpduData["prio%d" % (br.slotId-1,)], macStr)

    def SetRootTimes(self, ctp):
        """
        Copy the timers from the CTP packet into the root times to the ctpData
        """
        cd = self.ctpData
        cd.rootMsgAge    = socket.ntohs(ctp.age)
        cd.rootMaxAge    = socket.ntohs(ctp.max_age)
        cd.rootFwdDelay  = socket.ntohs(ctp.fd)
        cd.rootHelloTime = socket.ntohs(ctp.hello)

    def SelectMasterFC(self):
        """
        Evaluate to determine which fabric card should be the master of the I2C bus.
        The fabric cards are divided into four groups of four fabric cards. Each
        group selects a master fabric card for that group. If we have valid information
        from a card, if that card is the root fabric card or the first fabric card in
        the group, the card will be set as the master. Return True if the card
        we are running upon is a local master.
        """
        iWasLM = self.slotId in self.localMasters
        if self.cmdargs.chasStaticTopo:
            self.localMasters = [33, 37, 41, 45]
        else:
            self.localMasters = [0] * 4
            if self.ctpData is not None:
                for idx in xrange(33, 49):
                    group = (idx - 33) / 4
                    if self.ctpData.brExist[idx-1].age > 0:
                        if self.ctpData.rootPortId == idx or self.localMasters[group] == 0:
                            self.localMasters[group] = idx
        iAmLM = self.slotId in self.localMasters
        if iWasLM != iAmLM:
            if iAmLM: # Becoming local master
                self.InitTempData()
            else: # No longer local master
                self.tempData = None
        return iAmLM

    def SetStpDisabledState(self):
        """
        Place this card in the disabled state
        """
        self.ConfigAllPortsStp(bcm._PORT_CTRL_STP_STATE_DISABLED)

    def SetStpBlockingState(self):
        """
        Place this card in the blocking state
        """
        self.ConfigAllPortsStp(bcm._PORT_CTRL_STP_STATE_BLOCKING)

    def SetStpListeningState(self):
        """
        Place this card in the listening state
        """
        self.ConfigAllPortsStp(bcm._PORT_CTRL_STP_STATE_LISTENING)

    def SetStpForwardingState(self):
        """
        Place this card in the forwarding state. If the card is the root fabric
        card or the designated line card, set all ports to forwarding. Otherwise
        just put the port to the designated line card or root fabric card in
        forwarding and all other ports in blocking.
        """
        if (self.isFabricCard() and self.ctpData.rootPortId == self.slotId) or \
           (not self.isFabricCard() and self.ctpData.rootLCPortId == self.slotId):
            self.ConfigAllPortsStp(bcm._PORT_CTRL_STP_STATE_FORWARDING)
        else:
            if self.isFabricCard():
                fwdPort = "LC%d" % (((self.ctpData.rootLCPortId - 1) % 16) / 2,)
            else:
                fwdPort = "FC%d" % (((self.ctpData.rootPortId - 33) % 8) / 2,)

            for port in self.bcmPorts:
                if port.startswith("FC") or port.startswith("LC"):
                    if port == fwdPort:
                        self.ConfigPortStp(self.bcmPorts[port], bcm._PORT_CTRL_STP_STATE_FORWARDING)
                    else:
                        self.ConfigPortStp(self.bcmPorts[port], bcm._PORT_CTRL_STP_STATE_BLOCKING)

    def CardEnterDisable(self):
        """
        Place the card in the disabled state.
        """
        logger.debug("Card is entering the DISABLED state")
        self.ctpData.state = ctpInfo._STATE_DISABLE
        self.SetStpDisabledState()

    def CardEnterBlock(self):
        """
        Place this card in the blocking state. This is somewhat misleading because
        fabric cards in the blocking state have all of their ports forwarding.
        """
        logger.debug("Card is entering the BLOCKING state")
        self.ctpData.state = ctpInfo._STATE_BLOCK
        if self.isFabricCard():
            self.ClearDesignedLC()
            self.ConfigAllPortsStp(bcm._PORT_CTRL_STP_STATE_FORWARDING)
            self.SelectMasterFC()
        else:
            self.ClearRootFC()
            self.SetStpBlockingState()
        self.ctpData.timeSinceTopoChange = 0
        self.ClearBrExists()

    def CardEnterListening(self):
        """
        Place the card in the listening state.
        """
        logger.debug("Card is entering the LISTENING state")
        self.ctpData.state = ctpInfo._STATE_LISTENING
        if not self.isFabricCard():
            self.SetStpListeningState()
        self.ctpData.timeSinceTopoChange = 0

    def CardEnterLearning(self):
        """
        Place the card in the learning state.
        """
        logger.debug("Card is entering the LEARNING state")
        self.ctpData.state = ctpInfo._STATE_LEARNING
        if not self.isFabricCard():
            port = "FC%d" % (((self.ctpData.rootPortId - 33) % 8) / 2,)
            self.ConfigPortStp(self.bcmPorts[port], bcm._PORT_CTRL_STP_STATE_FORWARDING)
        self.ctpData.timeSinceTopoChange = 0

    def CardEnterForwarding(self):
        """
        Place the card in the learning state.
        """
        logger.debug("Card is entering the FORWARDING state")
        self.ctpData.state = ctpInfo._STATE_FORWARDING
        self.SetStpForwardingState()
        self.ctpData.timeSinceTopoChange = 0
        if self.isFabricCard():
            self.SelectMasterFC()

    def RunCtpStateMachine(self):
        """
        This function is periodically run to progress through the CTP state
        machine
        """
        # Clear the accounting of which CPDUs we have received in this interval
        self.rxCpduInterval = {}

        cd = self.ctpData
        if cd.state == ctpInfo._STATE_BLOCK:
            if self.isFabricCard():
                cd.timeSinceTopoChange += 1
                if cd.timeSinceTopoChange < self.cmdargs.chasMaxAge:
                    self.SendCpduPacket(ctpInfo._CPDU_TYPE_CONFIG)
                else:
                    cd.timeSinceTopoChange = 0
        elif cd.state == ctpInfo._STATE_LISTENING:
            if cd.timeSinceTopoChange < self.cmdargs.chasFwdDelay:
                cd.timeSinceTopoChange += 1
                if self.isFabricCard():
                    self.SendCpduPacket(ctpInfo._CPDU_TYPE_CONFIG)
            else:
                self.CardEnterLearning()
        elif cd.state == ctpInfo._STATE_LEARNING:
            fwdDelay = self.cmdargs.chasFwdDelay
            if not self.isFabricCard():
                fwdDelay = fwdDelay * 2 + 1
            if cd.timeSinceTopoChange < fwdDelay:
                self.SendCpduPacket(ctpInfo._CPDU_TYPE_CONFIG)
                cd.timeSinceTopoChange += 1
            else:
                self.CardEnterForwarding()
        elif cd.state == ctpInfo._STATE_FORWARDING:
            if self.slotId == cd.rootPortId:
                if cd.timeSinceTopoChange < self.cmdargs.chasMaxAge:
                    self.SendCpduPacket(ctpInfo._CPDU_TYPE_HELLO)
                    cd.timeSinceTopoChange += 1
                else:
                    if self.SelectNewRootLC(cd.rootLCPortId):
                        cd.timeSinceTopoChange = 0
                        self.SendCpduPacket(ctpInfo._CPDU_TYPE_HELLO)
                    else:
                        logger.debug("Entering block because of SelectNewRootLC failure")
                        self.CardEnterBlock()
                        self.SendCpduPacket(ctpInfo._CPDU_TYPE_LC_CHANGE)
            elif cd.timeSinceTopoChange > self.cmdargs.chasMaxAge*self.cmdargs.chasFwdDelay:
                logger.debug("Entering block because not rootPortId and long time since topo change")
                self.CardEnterBlock()
            else:
                cd.timeSinceTopoChange += 1

    def MacStrToByteStr(self, macStr):
        """
        Converts a MAC address string of the form "00:12:34:56:78:9a" to a 6
        byte string.
        """
        byteStr = macStr.split(":")
        result = ''
        for byte in byteStr:
            result += chr(int(byte, 16))
        return result

    def MacByteStrToStr(self, macByteStr):
        """
        Converts a string of 6 bytes to a string of the form "00:12:34:56:78:9a"
        """
        return ":".join(["%02x" % (ord(b),) for b in macByteStr])

    def BrIdToStr(self, brId):
        """
        Converts a bridge Id class (brId) to an 8-byte byte string.
        """
        addrStr = self.MacStrToByteStr(brId.addr)
        return struct.pack('!H', brId.prio) + addrStr

    def StrToBrId(self, brIdStr):
        """
        Converts an 8-byte byte string to a brId class
        """
        prio, addr = struct.unpack(">H6s", brIdStr)
        addrStr = self.MacByteStrToStr(addr)
        return brId(prio, addrStr)

    def CpduFlags(self, type):
        """
        Return the "flags" field value for a CPDU
        """
        flags = 0
        cd = self.ctpData

        if cd.state == ctpInfo._STATE_FORWARDING:
            flags |= ctpInfo._CPDU_FLAGS_LEARN
            flags |= ctpInfo._CPDU_FLAGS_FORWARD
        elif cd.state == ctpInfo._STATE_LEARNING:
            flags |= ctpInfo._CPDU_FLAGS_LEARN

        if type == ctpInfo._CPDU_TYPE_LC_CHANGE or type == ctpInfo._CPDU_TYPE_FC_CHANGE:
            flags |= ctpInfo._CPDU_FLAGS_TCN
        elif type == ctpInfo._CPDU_TYPE_CONFIG:
            flags |= ctpInfo._CPDU_FLAGS_PROPOSAL
        elif type == ctpInfo._CPDU_TYPE_REPLY:
            flags |= ctpInfo._CPDU_FLAGS_AGREEMENT
        elif type == ctpInfo._CPDU_TYPE_TCN:
            flags |= ctpInfo._CPDU_FLAGS_TCN_ACK

        if self.slotId == cd.rootBrId.prio & 0x00FF:
            flags |= ctpInfo._CPDU_FLAGS_ROOT_FC
        elif self.slotId == cd.designBrId.prio & 0x00FF:
            flags |= ctpInfo._CPDU_FLAGS_DESG_LC
        else:
            flags |= ctpInfo._CPDU_FLAGS_OTHER_ROLE

        return flags

    def SendCpduPacket(self, type):
        """
        Formulate and send a CPDU packet
        """
        # Build the packet from inner to outer. First, the CtpData
        self.DecBrExists()
        cd = self.ctpData
        kwargs={'ver1len' : 0 }
        for be in cd.brExist:
            kwargs["age%d" %  (be.slotId-1,)] = be.age
            kwargs["prio%d" % (be.slotId-1,)] = be.id.prio
            kwargs["addr%d" % (be.slotId-1,)] = self.MacStrToByteStr(be.id.addr)
        ctpD = CtpData(**kwargs)

        rootIdStr   = self.BrIdToStr(cd.rootBrId)
        bridgeIdStr = self.BrIdToStr(cd.designBrId)
        stpD = dpkt.stp.STP(proto_id=0, v=0, type=type, flags=self.CpduFlags(type), root_id=rootIdStr, 
                            root_path=cd.rootPathCost, bridge_id=bridgeIdStr, port_id=self.slotId,
                            age=socket.htons(cd.rootMsgAge), max_age=socket.htons(cd.rootMaxAge),
                            hello=socket.htons(cd.rootHelloTime), fd=socket.htons(cd.rootFwdDelay),
                            data=ctpD)
        llcD = LLC(dsap=0x42, ssap=0x42, ctrl=0x03, data=stpD)

        mcAddr = self.cmdargs.chasFcMcast if self.isFabricCard() else self.cmdargs.chasLcMcast
        if self.isLowerChassis():
            mcAddr = self.cmdargs.chasLcMcast if self.isFabricCard() else self.cmdargs.chasFcMcast
        dmacStr = self.MacStrToByteStr(mcAddr)
        smacStr = self.MacStrToByteStr(cd.brId.addr)
        eth = dpkt.ethernet.Ethernet(dst=dmacStr, src=smacStr, type=len(llcD), data=llcD)
        #self.DumpPacket(str(eth))
        if self.sock is not None:
            self.sock.send(str(eth))

    def SetDefaultTopology(self):
        """
        Programs the BMC53xx with the default topology.
        """
        self.ResetBcm53xx()
        if self.isFabricCard():
            # Fabric card initialization
            self.MakeMgmtVlanInternal()
            self.SetInitialPhy()
            modeReg = "switchMode96"
            sgmiiMode = bcm._PORT_SGMII_CNTRL_CD_EN      |\
                        bcm._PORT_SGMII_CNTRL_CRC_DIS    |\
                        bcm._PORT_SGMII_CNTRL_RSVD       |\
                        bcm._PORT_SGMII_CNTRL_AUTODET_EN |\
                        bcm._PORT_SGMII_CNTRL_FIBER
        else:
            # Line card initialization
            modeReg = "switchMode89"
            sgmiiMode = bcm._PORT_SGMII_CNTRL_CD_EN      |\
                        bcm._PORT_SGMII_CNTRL_CRC_DIS    |\
                        bcm._PORT_SGMII_CNTRL_RSVD       |\
                        bcm._PORT_SGMII_CNTRL_FIBER

        self.SetSGMIIMode(sgmiiMode)
        self.WriteBcm53xx(modeReg,
                          bcm._SWITCH_MODE_RTRY_LMT_DIS |\
                          bcm._SWITCH_MODE_SW_FWDG_EN   |\
                          bcm._SWITCH_MODE_SW_FWDG_MODE)
        self.SetStaticARL()
        self.SetReservedMulticastFwd()
        self.InitialPortStpState()

    def ClearDefaultTopology(self):
        """
        Breaks down the default topology so that loops can't form.
        """
        if self.isCpuA() and not self.cmdargs.chasStaticTopo and self.ctpData is not None:
            self.CardEnterBlock()
            if self.slotId == self.ctpData.rootLCPortId:
                self.SendCpduPacket(ctpInfo._CPDU_TYPE_FC_CHANGE)

    def MacStrToInt(self, macStr):
        """
        Converts a MAC address in colon-separated string to an integer.
        """
        byteStr = macStr.split(":")
        result = 0
        for byte in byteStr:
            result = (result << 8) + int(byte, 16)
        return result

    def GenBpfFilter(self):
        """
        Generate the BPF filters and program from tcpdump output.
        """

        _BPF_INS_IDX_CODE = 0
        _BPF_INS_IDX_JT   = 1
        _BPF_INS_IDX_JF   = 2
        _BPF_INS_IDX_K    = 3

        # Output of: tcpdump -dd ether dst 01:10:b5:00:00:01 and inbound
        bpfInstr = [ [ 0x20, 0, 0, 0x00000002 ],
                     [ 0x15, 0, 5, 0xb5000001 ],
                     [ 0x28, 0, 0, 0x00000000 ],
                     [ 0x15, 0, 3, 0x00000110 ],
                     [ 0x28, 0, 0, 0xfffff004 ],
                     [ 0x15, 1, 0, 0x00000004 ],
                     [ 0x6,  0, 0, 0x00040000 ],
                     [ 0x6,  0, 0, 0x00000000 ] ]

        # Fabric cards RECEIVE the line card's multicast, except on lower chassis
        recvMac = self.cmdargs.chasLcMcast if self.isFabricCard() else self.cmdargs.chasFcMcast
        if self.isLowerChassis():
            recvMac = self.cmdargs.chasFcMcast if self.isFabricCard() else self.cmdargs.chasLcMcast
        recvMacInt = self.MacStrToInt(recvMac)

        # WARN: This code is dependent on the BPF instructions in bpfInstr
        # Set the MAC address we want to receive and vlan id
        bpfInstr[1][_BPF_INS_IDX_K] = recvMacInt & 0xFFFFFFFF
        bpfInstr[3][_BPF_INS_IDX_K] = (recvMacInt >> 32) & 0xFFFFFFFF

        filter_array = struct_sock_filter * len(bpfInstr)
        self.filters = filter_array()
        idx = 0
        for line in bpfInstr:
            self.filters[idx].code = line[_BPF_INS_IDX_CODE]
            self.filters[idx].jt = line[_BPF_INS_IDX_JT]
            self.filters[idx].jf = line[_BPF_INS_IDX_JF]
            self.filters[idx].k = line[_BPF_INS_IDX_K]
            idx += 1

        fprog = struct_sock_fprog()
        fprog.len = len(bpfInstr)
        fprog.filter = ctypes.pointer(self.filters[0])
        self.fprog = ctypes.string_at(ctypes.addressof(fprog), ctypes.sizeof(fprog))
        return

    def InitSelfPipe(self):
        """
        Create a pipe to ourselves. The only purpose of this pipe is so that it
        can be added to the read object in the select.select() call so that a
        write to this pipe, will cause the select.select() call to return. This
        is only used when exiting the program.
        """
        try:
            SelfPipeR, SelfPipeW = os.pipe()
        except os.error as e:
            logger.error("Could not open self pipe (%s)" % (str(e),))
            return None, None

        try:
            flags = fcntl.fcntl(SelfPipeR, fcntl.F_GETFL)
            fcntl.fcntl(SelfPipeR, fcntl.F_SETFL, flags | os.O_NONBLOCK)
            flags = fcntl.fcntl(SelfPipeW, fcntl.F_GETFL)
            fcntl.fcntl(SelfPipeW, fcntl.F_SETFL, flags | os.O_NONBLOCK)
        except IOError as e:
            logger.error("Could not get or set flags on self pipe (%s)" % (str(e),))
            os.close(SelfPipeR)
            os.close(SelfPipeW)
            return None, None
        return SelfPipeR, SelfPipeW

    def isOperUp(self, ifName):
        """
        Return True if the supplied interface is oper UP.
        """
        try:
            with open("/sys/class/net/" + ifName + "/operstate") as f:
                operstate = f.readline().strip()
        except (AttributeError, IOError):
            operstate = None
        return operstate == "up"

    def GetIfMac(self, ifName):
        """
        Get the MAC address for the supplied interface name
        """
        try:
            with open("/sys/class/net/" + ifName + "/address") as f:
                addr = f.readline().strip()
        except (AttributeError, IOError):
            addr = None
        return addr

    def UpMgmtIfs(self):
        """
        Bring up the management interfaces
        """
        vlanIfName = self.cmdargs.chasIfName + ".%d" % (self.cmdargs.chasVlan,)
        if not self.isOperUp(vlanIfName):
            try:
                subprocess.check_call(["/sbin/ifup", self.cmdargs.chasIfName, "--admin-state"])
                subprocess.check_call(["/sbin/ifup", vlanIfName])
            except (subprocess.CalledProcessError, OSError):
                self.ctpData = None
                self.sock = None

    def CtpRx(self):
        """
        The CTP Receive Thread
        """
        logger.debug("Beginning execution of the CtpRx thread.")

        while not self.stopEvent.is_set():

            # Create the socket
            while (self.sock is None or self.ctpData is None) and not self.stopEvent.is_set():

                vlanIfName = self.cmdargs.chasIfName + ".%d" % (self.cmdargs.chasVlan,)
                if self.ctpData is None:
                    addr = self.GetIfMac(vlanIfName)
                    if addr is not None:
                        self.InitCtpData(addr)

                # create a socket for all ethertypes - 0x3
                if self.sock is None:
                    sock = socket.socket(socket.PF_PACKET, socket.SOCK_RAW, socket.htons(_ETH_P_ALL))

                    try:
                        # bind socket to management interface
                        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                        sock.bind((vlanIfName, 0))

                        # attach bpf filter to only get arp packets
                        sock.setsockopt(socket.SOL_SOCKET, _SO_ATTACH_FILTER, self.fprog)
                        self.sock = sock

                    except socket.error:
                        sock.close()
                        self.sock = None

                if self.ctpData is None or self.sock is None:
                    time.sleep(1)
                    self.UpMgmtIfs()

            # Read data until the socket is closed
            while self.sock is not None and not self.stopEvent.is_set():

                inputs = [self.selfPipeR, self.sock]
                try:
                    (readable, writable, exceptional) = select.select(inputs, [], inputs)
                except Exception as e:
                    logger.info('select error: %r' % e)
                    readable = writable = exceptional = []

                for errSock in exceptional:
                    logger.warning('exceptSock: %r exception ignored' % errSock)

                if self.selfPipeR in readable:
                    logger.debug("Received data on self pipe. Preparing to exit thread.")
                    while True:
                        try:
                            os.read(self.selfPipeR, 1)
                        except OSError as e:
                            # Is the pipe totally drained? Exit.
                            if e.errno != errno.EAGAIN:
                                logger.error("An error occured while draining the self pipe: %s" % (str(e),))
                            break

                if self.sock in readable:
                    self.RxCtpMsg(self.sock)

        logger.debug("Execution of the CtpRx thread is complete.")

    def RxCtpMsg(self, sock):
        """
        This function is invoked to read in and process a received CTP message.
        """
        try:
            recvData = sock.recv(4096)
        except socket.error as e:
            logger.warning("Unable to receive data from the CTP socket: %s" % (str(e),))
            sock.close()
            self.sock = None
            return

        #self.DumpPacket(recvData)
        try:
            eth = dpkt.ethernet.Ethernet(recvData)
        except:
            return

        if not isinstance(eth.data, dpkt.stp.STP):
            return
        ctp = eth.data

        # Have we already received this type of message from this sender since
        # we sent a CPDU
        bridgeSlotId = ctp.port_id
        if bridgeSlotId in self.rxCpduInterval.get(ctp.type, set()):
            return
        self.rxCpduInterval.setdefault(ctp.type, set()).add(bridgeSlotId)

        if self.isFabricCard(bridgeSlotId):
            self.RxFCCtpMsg(ctp)
        else:
            self.RxLCCtpMsg(ctp)

    def RxFCCtpMsg(self, ctp):
        """
        Process a CTP message received from a fabric card
        """
        # Only process messages from fabric cards on line cards
        cd = self.ctpData
        if not self.isFabricCard():
            if len(ctp.data) < CtpData.__hdr_len__:
                return
            cpduData = CtpData(ctp.data)
            if cd.state == ctpInfo._STATE_BLOCK:
                self.CardEnterListening()
                cd.topoChange          = True
                cd.timeSinceTopoChange = 0
                cd.topoChangeCount     = 1
                self.UpdateLCRoot(ctp)
                self.SetRootTimes(ctp)
                self.SendCpduPacket(ctpInfo._CPDU_TYPE_REPLY)
                self.UpdateBrExists(cpduData)
            elif cd.state == ctpInfo._STATE_LISTENING or cd.state == ctpInfo._STATE_LEARNING:
                if self.CheckCpduNewIds(ctp):
                    self.UpdateLCRoot(ctp)
                self.UpdateBrExists(cpduData)
            elif cd.state == ctpInfo._STATE_FORWARDING:
                self.UpdateBrExists(cpduData)
                if ctp.type == ctpInfo._CPDU_TYPE_LC_CHANGE:
                    self.SendCpduPacket(ctpInfo._CPDU_TYPE_LC_CHANGE)
                    cd.timeSinceTopoChange = 0
                    origDBridgeSlotId = cd.brId.prio & 0x00FF
                    if (cd.rootLCPortId != origDBridgeSlotId) and (cd.brExist[origDBridgeSlotId-1].age > 0):
                        cd.brExist[origDBridgeSlotId-1].age = 0
                    self.CardEnterBlock()
                elif ctp.type == ctpInfo._CPDU_TYPE_HELLO:
                    cd.timeSinceTopoChange = 0
                    rootSlotId        = self.StrToBrId(ctp.root_id).prio & 0x00FF
                    dBridgeSlotId     = self.StrToBrId(ctp.bridge_id).prio & 0x00FF
                    origRootSlotId    = cd.rootBrId.prio & 0x00FF
                    origDBridgeSlotId = cd.brId.prio & 0x00FF
                    if rootSlotId == origRootSlotId and dBridgeSlotId != cd.rootLCPortId:
                        self.LCSyncToHello(ctp)
                    elif dBridgeSlotId == origDBridgeSlotId and rootSlotId == origRootSlotId and ctp.port_id == rootSlotId:
                        self.SendCpduPacket(ctpInfo._CPDU_TYPE_HELLO)
                    else:    
                        self.SendCpduPacket(ctpInfo._CPDU_TYPE_REPLY)
                elif ctp.type == ctpInfo._CPDU_TYPE_CONFIG:
                    self.SendCpduPacket(ctpInfo._CPDU_TYPE_REPLY)

    def RxLCCtpMsg(self, ctp):
        """
        Process a CTP message received from a line card
        """
        # Only process messages from line cards on fabric cards
        cd = self.ctpData
        if self.isFabricCard():
            if len(ctp.data) < CtpData.__hdr_len__:
                return
            cpduData = CtpData(ctp.data)
            if cd.state == ctpInfo._STATE_BLOCK:
                self.CardEnterListening()
                cd.topoChange          = True
                cd.timeSinceTopoChange = 0
                cd.topoChangeCount     = 1
                self.UpdateFCDesignated(ctp)
                self.SetRootTimes(ctp)
                self.UpdateBrExists(cpduData)
            elif cd.state == ctpInfo._STATE_LISTENING or cd.state == ctpInfo._STATE_LEARNING:
                if self.CheckCpduNewIds(ctp):
                    self.UpdateFCDesignated(ctp)
                self.UpdateBrExists(cpduData)
            elif cd.state == ctpInfo._STATE_FORWARDING:
                self.UpdateBrExists(cpduData)
                if ctp.type == ctpInfo._CPDU_TYPE_HELLO:
                    cd.timeSinceTopoChange = 0
                    rootSlotId        = self.StrToBrId(ctp.root_id).prio & 0x00FF
                    dBridgeSlotId     = self.StrToBrId(ctp.bridge_id).prio & 0x00FF
                    origRootSlotId    = cd.rootBrId.prio & 0x00FF
                    origDBridgeSlotId = cd.designBrId.prio & 0x00FF
                    myId              = cd.brId.prio & 0x00FF
                    if dBridgeSlotId != origDBridgeSlotId and rootSlotId == origRootSlotId:
                        self.LCSyncToHello(ctp)
                    elif dBridgeSlotId == origDBridgeSlotId and rootSlotId != myId:
                        self.SendCpduPacket(ctpInfo._CPDU_TYPE_HELLO)
                elif ctp.type == ctpInfo._CPDU_TYPE_CONFIG:
                    self.SendCpduPacket(ctpInfo._CPDU_TYPE_REPLY)
                elif ctp.type == ctpInfo._CPDU_TYPE_FC_CHANGE:
                    rootSlotId = self.StrToBrId(ctp.root_id).prio & 0x00FF
                    myId       = cd.brId.prio & 0x00FF
                    if rootSlotId == myId:
                        oldRootLC = self.StrToBrId(ctp.bridge_id).prio & 0x00FF
                        if SelectNewRootLC(oldRootLC):
                            self.SendCpduPacket(ctpInfo._CPDU_TYPE_HELLO)
                        else:
                            self.CardEnterBlock()
                elif ctp.type == ctpInfo._CPDU_TYPE_LC_CHANGE:
                    rootSlotId = self.StrToBrId(ctp.root_id).prio & 0x00FF
                    myId       = cd.brId.prio & 0x00FF
                    if rootSlotId != myId:
                        self.CardEnterBlock()

    def DumpPacket(self, packet):
        """
        Dump the contents of a packet to the debug log in a nice format.
        """
        index = 0
        outline = "Packet (len=%d, 0x%04x):" % (len(packet), len(packet))
        for byte in packet:
            if index % 16 == 0:
                logger.debug(outline)
                outline= "  %04x : " % (index,)
            outline += "%02x " % (ord(byte),)
            if index % 8 == 7:
                outline += " "
            index += 1
        logger.debug(outline)

    def CtpTx(self):
        """
        The CTP Transmit Thread
        """
        logger.debug("Beginning execution of the CtpTx thread.")

        procStart = os.times()[4]
        while not self.stopEvent.is_set():

            # Do the work of the thread here
            if self.ctpData is not None and self.ctpData.adminState == ctpInfo._CTP_ENABLED:
                self.RunCtpStateMachine()

            # Wake up every interval, factor in the time to do processing
            sleepTime = (self.cmdargs.chasHelloTime / 1000.0) - (os.times()[4] - procStart)
            if sleepTime > 0:
                self.stopEvent.wait(sleepTime)
            procStart = os.times()[4]

        logger.debug("Execution of the CtpTx thread is complete.")

    def GetCpuTemp(self):
        """
        Return the maximum temperature of the CPU temp sensors in LM75 format.
        """
        maxTemp = -127
        tempNodes = glob.glob("/sys/devices/platform/coretemp.*/hwmon/hwmon*/temp*_input")
        for t in tempNodes:
            try:
                with open(t) as f:
                    currTemp = (int(f.readline().strip()) + 500) / 1000
            except (AttributeError, IOError, ValueError):
                currTemp = 127
            maxTemp = currTemp if currTemp > maxTemp else maxTemp
        # Convert to 2's complement 8-bit hex string
        maxTemp = 127 if maxTemp > 127 else maxTemp
        maxTemp = -128 if maxTemp < -128 else maxTemp
        return hex(maxTemp & 0xff)

    def GetTomahawkTemp(self):
        """
        Return the maximum temperature of the Tomahawk temp sensors in LM75 format.
        """
        _TEMP_PATH = "/cumulus/switchd/run/"

        # How many temp sensors are there?
        maxTemp = -127
        try:
            with open(_TEMP_PATH + "Temp_Max") as f:
                numTemps = int(f.readline().strip())
        except (AttributeError, IOError, ValueError):
            numTemps = 0
            maxTemp = 127

        # Read each one and find the maximum
        for t in xrange(numTemps):
            try:
                with open(_TEMP_PATH + "temp%d_input" % (t+1,)) as f:
                    currTemp = (int(f.readline().strip()) + 500) / 1000
            except (AttributeError, IOError, ValueError):
                currTemp = 127
            maxTemp = currTemp if currTemp > maxTemp else maxTemp

        # Convert to 2's complement 8-bit hex string
        maxTemp = 127 if maxTemp > 127 else maxTemp
        maxTemp = -128 if maxTemp < -128 else maxTemp
        return hex(maxTemp & 0xff)

    def UpdateLocalTemps(self):
        """
        Update the CPLD1 registers with the temps of the Tomahawk and CPU so that
        the FANCPLD can use these temps in its calculations.
        """
        self.WriteCpld1("cpu_thermal", self.GetCpuTemp())
        self.WriteCpld1("mac_thermal", self.GetTomahawkTemp())

    def ReadFanController(self, idx, node):
        """
        Read a value from the fan controller
        """
        _I2C_PATH = "/sys/bus/i2c/devices/"
        fanCtrl = "10-0066/" if idx == 0 else "16-0066/"
        sensorPath = _I2C_PATH + fanCtrl + node
        return self.ReadSysfsNode(sensorPath)

    def WriteFanController(self, idx, node, value):
        """
        Write a value to the fan controller
        """
        _I2C_PATH = "/sys/bus/i2c/devices/"
        fanCtrl = "10-0066/" if idx == 0 else "16-0066/"
        sensorPath = _I2C_PATH + fanCtrl + node
        self.WriteSysfsNode(sensorPath, value)

    def GetSensorTemp(self, idx, card, sensor):
        """
        Read a temperature sensor's temperature
        """
        node = card + "_" + sensor + "_temp"
        tempStr = self.ReadFanController(idx, node)
        try:
            temp = int(tempStr, 16)
            temp = temp - int((temp << 1) & 0x100)
        except (TypeError, ValueError):
            temp = None
        return temp

    def UpdateLocalMasterTemps(self):
        """
        Update the temperatures maintained by this local master fabric card.
        """
        for card, fanCntrlList in vars(self.tempData).iteritems():
            for fanCtrl in fanCntrlList:
                for sensor, tempInfo in vars(fanCtrl).iteritems():
                    if sensor != "idx":
                        currTemp = self.GetSensorTemp(fanCtrl.idx, card, sensor)
                        if currTemp is None:
                            tempInfo.errors += 1
                        else:
                            # CPU hasn't had chance to update CPU/ASIC temp
                            if sensor in ['asic', 'cpu'] and currTemp == 127:
                                currTemp = 63
                            tempInfo.temp = currTemp
                            tempInfo.errors = 0

    def CardNameToSlotId(self, name):
        """
        Converts a card name to a slot ID.
        """
        nameToSlotId = { "lc101"  :  1, "lc102"  :  2, "lc201"  :  3, "lc202"  :  4,
                         "lc301"  :  5, "lc302"  :  6, "lc401"  :  7, "lc402"  :  8,
                         "lc501"  :  9, "lc502"  : 10, "lc601"  : 11, "lc602"  : 12,
                         "lc701"  : 13, "lc702"  : 14, "lc801"  : 15, "lc802"  : 16,
                         "lc901"  : 17, "lc902"  : 18, "lc1001" : 19, "lc1002" : 20,
                         "lc1101" : 21, "lc1102" : 22, "lc1201" : 23, "lc1202" : 24,
                         "lc1301" : 25, "lc1302" : 26, "lc1401" : 27, "lc1402" : 28,
                         "lc1501" : 29, "lc1502" : 30, "lc1601" : 31, "lc1602" : 32,
                         "fc101"  : 33, "fc102"  : 34, "fc201"  : 35, "fc202"  : 36,
                         "fc301"  : 37, "fc302"  : 38, "fc401"  : 39, "fc402"  : 40,
                         "fc501"  : 41, "fc502"  : 42, "fc601"  : 43, "fc602"  : 44,
                         "fc701"  : 45, "fc702"  : 46, "fc801"  : 47, "fc802"  : 48 }
        return nameToSlotId.get(name.lower(), 0)

    def FanCtrlToCardName(self, node, fanCtrl):
        """
        Converts the fan controller's node (fca) to a card name (fc302).
        """
        fanCtrlToCard = {}
        if self.slotId in [33, 35]:
            fanCtrlToCard = { (0, "fca") : "fc102",  (0, "fcb") : "fc202",  (0, "lca") : "lc501",
                              (0, "lcb") : "lc601",  (0, "lcc") : "lc701",  (0, "lcd") : "lc801",
                              (1, "fca") : "fc101",  (1, "fcb") : "fc201",  (1, "lca") : "lc101",
                              (1, "lcb") : "lc201",  (1, "lcc") : "lc301",  (1, "lcd") : "lc401" }
        elif self.slotId in [37, 39]:
            fanCtrlToCard = { (0, "fca") : "fc302",  (0, "fcb") : "fc402",  (0, "lca") : "lc102",
                              (0, "lcb") : "lc202",  (0, "lcc") : "lc302",  (0, "lcd") : "lc402",
                              (1, "fca") : "fc301",  (1, "fcb") : "fc401",  (1, "lca") : "lc502",
                              (1, "lcb") : "lc602",  (1, "lcc") : "lc702",  (1, "lcd") : "lc802" }
        elif self.slotId in [41, 43]:                                              
            fanCtrlToCard = { (0, "fca") : "fc502",  (0, "fcb") : "fc602",  (0, "lca") : "lc1301",
                              (0, "lcb") : "lc1401", (0, "lcc") : "lc1501", (0, "lcd") : "lc1601",
                              (1, "fca") : "fc501",  (1, "fcb") : "fc601",  (1, "lca") : "lc901",
                              (1, "lcb") : "lc1001", (1, "lcc") : "lc1101", (1, "lcd") : "lc1201" }
        elif self.slotId in [45, 47]:                                              
            fanCtrlToCard = { (0, "fca") : "fc702",  (0, "fcb") : "fc802",  (0, "lca") : "lc902",
                              (0, "lcb") : "lc1002", (0, "lcc") : "lc1102", (0, "lcd") : "lc1202",
                              (1, "fca") : "fc701",  (1, "fcb") : "fc801",  (1, "lca") : "lc1302",
                              (1, "lcb") : "lc1402", (1, "lcc") : "lc1502", (1, "lcd") : "lc1602" }
        return fanCtrlToCard.get((fanCtrl, node), "?????")

    def HandleTempWarnings(self):
        """
        Find any temperature sensors which are above the warning threshold
        """
        for card, fanCntrlList in vars(self.tempData).iteritems():
            for fanCtrl in fanCntrlList:
                for sensor, tempInfo in vars(fanCtrl).iteritems():
                    if sensor != "idx":
                        if tempInfo.temp > tempInfo.warn:
                            cardName = self.FanCtrlToCardName(card, fanCtrl.idx)
                            logger.warning("The %s temperature of sensor on card %s is %d C, which is above the warning threshold %d C." % (sensor, cardName, tempInfo.temp, tempInfo.warn))

    def ShutdownCard(self, slotId, slotName):
        """
        Shuts down the card with the supplied slot ID
        """
        _I2C_PATH = "/sys/bus/i2c/devices/"
        admNodes = ["14-0010", "14-0013", "14-0050", "14-0053", "14-0010", "14-0013", "14-0050", "14-0053",
                    "14-0010", "14-0013", "14-0050", "14-0053", "14-0010", "14-0013", "14-0050", "14-0053",
                    "14-0044", "14-0047", "14-0044", "14-0047", "14-0044", "14-0047", "14-0044", "14-0047"]
        canControl = (self.slotId in [33, 35] and (1  <= slotId <= 8  or 33 <= slotId <= 36)) or \
                     (self.slotId in [37, 39] and (9  <= slotId <= 16 or 37 <= slotId <= 40)) or \
                     (self.slotId in [41, 43] and (17 <= slotId <= 24 or 41 <= slotId <= 44)) or \
                     (self.slotId in [45, 47] and (25 <= slotId <= 32 or 45 <= slotId <= 48))

        if self.cmdargs.chasNoShutdown:
            logger.debug("Card %s in slotId %d is not being shut down because --chasNoShutdown is enabled." % (clotName, slotId))
        elif not canControl:
            logger.critical("Cannot power down cardId %d (%s) from cardId %d. This card is at risk of damage!" % (slotId, slotName, self.slotId))
        else:
            admNode = admNodes[(slotId-1)/2]
            hsePath = _I2C_PATH + admNode + "/hot_swap_enable"
            logger.critical("POWERING DOWN CARD %s!!!" % (slotName,))
            self.WriteSysfsNode(hsePath, "0")

    def HandleTempShutdown(self):
        """
        Find any temperature sensors which are above the shutdown threshold
        """
        shutdown = {}
        for card, fanCntrlList in vars(self.tempData).iteritems():
            for fanCtrl in fanCntrlList:
                for sensor, tempInfo in vars(fanCtrl).iteritems():
                    if sensor != "idx":
                        if tempInfo.temp > tempInfo.shutdown:
                            cardName = self.FanCtrlToCardName(card, fanCtrl.idx)
                            logger.critical("The %s temperature of sensor on card %s is %d C, which is above the shutdown threshold %d C." % (sensor, cardName, tempInfo.temp, tempInfo.shutdown))
                            shutdown[self.CardNameToSlotId(cardName)] = cardName

        shutdownSelf = None
        for slotId in shutdown:
            # Dont' shut ourselves down until the other cards are off first
            slotName = shutdown.get(slotId,"???")
            if (slotId-1)/2 == (self.slotId-1)/2:
                shutdownSelf = slotName
            else:
                self.ShutdownCard(slotId, slotName)

        # Suicide
        if shutdownSelf:
            self.ShutdownCard(self.slotId, shutdownSelf)

    def Thermal(self):
        """
        The thermal manager thread. This runs on both the FC and LC and both
        CPU-A and CPU-B.
        """
        logger.debug("Beginning execution of the thermal thread.")

        procStart = thermStart = os.times()[4]
        while not self.stopEvent.is_set():

            # Enable and tickle the watchdog timer
            if self.SelectMasterFC():
                self.WriteFanController(0, "fan_watchdog_enable", 1)
                self.WriteFanController(1, "fan_watchdog_enable", 1)
                self.WriteFanController(0, "fan_watchdog", 6)
                self.WriteFanController(1, "fan_watchdog", 6)

            # The thermal manager runs less frequently
            if self.cmdargs.chasThermalFreq - (os.times()[4] - thermStart) <= 0:
                thermStart = os.times()[4]
                self.UpdateLocalTemps()
                if self.SelectMasterFC():
                    self.UpdateLocalMasterTemps()
                    self.HandleTempWarnings()
                    self.HandleTempShutdown()

            # Wake up every interval, factor in the time to do processing
            sleepTime = 1.0 - (os.times()[4] - procStart)
            if sleepTime > 0:
                self.stopEvent.wait(sleepTime)
            procStart = os.times()[4]

        logger.debug("Execution of the thermal thread is complete.")

    def DumpTraceback(self):
        """
        Dump a nicely formatted traceback to the log file
        """
        (exc_type, exc_value, exc_traceback) = sys.exc_info()
        err = "".join(traceback.format_exception(exc_type, exc_value,
                                                 exc_traceback))
        logger.error("unhandled exception:")
        for line in err.split('\n'):
            logger.error("    %s" % (line,))

    def CtpRxT(self):
        try:
            self.CtpRx()
        except Exception as e:
            self.DumpTraceback()
            self.stopEvent.set()

    def CtpTxT(self):
        try:
            self.CtpTx()
        except Exception as e:
            self.DumpTraceback()
            self.stopEvent.set()

    def ThermalT(self):
        try:
            self.Thermal()
        except Exception as e:
            self.DumpTraceback()
            self.stopEvent.set()

if __name__ == '__main__':
    pass
