#! /usr/bin/python

#-------------------------------------------------------------------------------
#
#   Copyright (C) 2017 Cumulus Networks, Inc. All rights reserved
#
#   This software is subject to the Cumulus Networks End User License
#   Agreement available  at the following locations:
#
#   Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#   Cumulus Linux systems: /usr/share/cumulus/EULA.txt
#
#-------------------------------------------------------------------------------

"""
This module implements the command line interface server for the cumulus
chassisd service.
"""


#-------------------------------------------------------------------------------
#
#   Import the necessary modules
#
#-------------------------------------------------------------------------------

try:
    import argparse
    import errno
    import fcntl
    import json
    import logging
    import os
    import platform
    import select
    import socket
    import struct
    import sys
    import threading
    import traceback
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")


#-------------------------------------------------------------------------------
#
#   Global Variables
#
#-------------------------------------------------------------------------------

# The handle used for logging messages
logger = logging.getLogger("cumulus-chassisd")


#-------------------------------------------------------------------------------
#
#   The command line server
#
#-------------------------------------------------------------------------------

class chassisd_cmdsrv:
    """
    This class handles everything required to implement the command server
    interface for the chassisd service. This interface allows other processes to
    communicate with chassisd to get status and alter configuration. chassisctl
    is an example of a process which connects to this interface.
    """

    _CmdSocketFile = "/var/run/chassisd.socket"

    #   XXX: This is a hack. Python socket module doesn't define the SO_PEERCRED
    #   value. And this value is different on different CPU architectures. So on
    #   PowerPC platforms the value is 21 and on all other platforms the value is 17.
    _SOCKET_PEERCRED = 21 if platform.machine() == "ppc" else 17
    _DEFAULT_JSON_INDENT = 4

    def __init__(self, chassis, cmdline, stopEvent, actchd):
        # Save some command line parameters which we will use later
        self.chassis        = chassis
        self.cmdargs, self.cmdline = self.ParseCmdLine(cmdline)
        self.stopEvent      = stopEvent
        self.actchd         = actchd

        #
        #   Command interface socket. The cmdSrvSock is a socket to which clients
        #   can connect to open up communication with the chassisd daemon. The cmdOutputBufs
        #   is a dictionary keyed by the client socket, with the string of data to write
        #   to the socket as the value. Only sockets which have data waiting to be output 
        #   are in this dictionary.
        #
        self.cmdSrvSock         = None
        self.cmdOutputLock      = threading.Lock()
        self.cmdOutputBufs      = {}

        self.selfPipeR, self.selfPipeW = self.InitSelfPipe()
        if self.selfPipeR is None or self.selfPipeW is None:
            logger.error("Could not create the self pipe.")
            self.stopEvent.set()
            return


        #
        #   Start a thread to handle requests from other tools
        #
        self.commandThread = threading.Thread(None, self.CmdRecvT, "CmdRecv")
        self.commandThread.daemon = True
        self.commandThread.start()

    def __del__(self):
        """
        Destructor:
        """
        if self.cmdSrvSock is not None:
            cmdSrvSock = self.cmdSrvSock
            self.cmdSrvSock = None
            cmdSrvSock.shutdown(socket.SHUT_RDWR)

    def ParseCmdLine(self, cmdline):
        parser = argparse.ArgumentParser(description="The Command Line Server")
        parser.add_argument("--cmdNoSort", action="store_false",
                            help="No not sort the JSON send to the client.")
        parser.add_argument("--cmdIndent", default=chassisd_cmdsrv._DEFAULT_JSON_INDENT,
                            metavar="SPACES", help="The amount to indent each JSON level.")
        args, cmdline = parser.parse_known_args(cmdline)
        return(args, cmdline)

    def SignalThreads(self):
        if self.selfPipeW is not None:
            try:
                os.write(self.selfPipeW, 'x')
            except os.error:
                pass

    def JoinThreads(self):
        if self.commandThread is not None and self.commandThread.is_alive():
            self.commandThread.join()

    def InitSelfPipe(self):
        """
        Create a pipe to ourselves. The only purpose of this pipe is so that it
        can be added to the read object in the select.select() call so that a
        write to this pipe, will cause the select.select() call to return. This
        is only used when exiting the program.
        """
        try:
            SelfPipeR, SelfPipeW = os.pipe()
        except os.error as e:
            logger.error("Could not open self pipe (%s)" % (str(e),))
            return None, None

        try:
            flags = fcntl.fcntl(SelfPipeR, fcntl.F_GETFL)
            fcntl.fcntl(SelfPipeR, fcntl.F_SETFL, flags | os.O_NONBLOCK)
            flags = fcntl.fcntl(SelfPipeW, fcntl.F_GETFL)
            fcntl.fcntl(SelfPipeW, fcntl.F_SETFL, flags | os.O_NONBLOCK)
        except IOError as e:
            logger.error("Could not get or set flags on self pipe (%s)" % (str(e),))
            os.close(SelfPipeR)
            os.close(SelfPipeW)
            return None, None
        return SelfPipeR, SelfPipeW

    def DumpTraceback(self):
        """
        Dump a nicely formatted traceback to the log file
        """
        (exc_type, exc_value, exc_traceback) = sys.exc_info()
        err = "".join(traceback.format_exception(exc_type, exc_value,
                                                 exc_traceback))
        logger.error("unhandled exception:")
        for line in err.split('\n'):
            logger.error("    %s" % (line,))

    def CmdRecvT(self):
        try:
            self.CmdRecv()
        except Exception as e:
            self.DumpTraceback()
            self.stopEvent.set()

    def CmdRecv(self):
        '''
        This function is an independent thread of the script. It creates a socket,
        cmdSrvSock, and accepts connections on that socket. It receives the commands
        sent through those connections and replies with the proper command responses.
        '''
        logger.debug("Beginning execution of the thread CmdRecv")

        # Create a socket upon which we will accept connections
        try:
            self.cmdSrvSock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.cmdSrvSock.setblocking(0)
        except socket.error as msg:
            self.cmdSrvSock = None
            logger.debug("Unable to create a Unix stream socket (%s)" % (msg,))
        else:
            try:
                os.remove(chassisd_cmdsrv._CmdSocketFile)
            except OSError:
                pass

            try:
                self.cmdSrvSock.bind(chassisd_cmdsrv._CmdSocketFile)
                os.chmod(chassisd_cmdsrv._CmdSocketFile, 0777)
                self.cmdSrvSock.listen(5)
                logger.debug("Server socket to %s was created" % (chassisd_cmdsrv._CmdSocketFile))
            except socket.error as msg:
                self.cmdSrvSock.close()
                os.remove(chassisd_cmdsrv._CmdSocketFile)
                self.cmdSrvSock = None
                logger.debug("Unable to bind socket to %s (%s)" % (chassisd_cmdsrv._CmdSocketFile, msg))

        inputBufs = {}
        inputs  = [ self.cmdSrvSock ]
        while (inputs and self.cmdSrvSock and not self.stopEvent.is_set()) or self.cmdOutputBufs:

            self.cmdOutputLock.acquire()
            cmdOutputSocks = self.cmdOutputBufs.keys()
            self.cmdOutputLock.release()

            # Wait until there is a socket to read, write, or an error
            try:
                (readable, writable, exceptional) = select.select(inputs + [self.selfPipeR], cmdOutputSocks, inputs + [self.selfPipeR] + cmdOutputSocks)
            except select.error as e:
                logger.debug("A select.error occurred")
                (readable, writable, exceptional) = ([],[],[])
            except socket.error:
                logger.debug("A socket.error occurred")
                (readable, writable, exceptional) = ([],[],[])

            for readSock in readable:
                if readSock is self.selfPipeR:
                    logger.debug("Received data on self pipe. Preparing to exit thread.")
                    while True:
                        try:
                            os.read(self.selfPipeR, 1)
                        except OSError as e:
                            # Is the pipe totally drained? Exit.
                            if e.errno != errno.EAGAIN:
                                logger.error("An error occured while draining the self pipe: %s" % (str(e),))
                            break
                elif readSock is self.cmdSrvSock:
                    # Accept a new socket connection
                    logger.debug("Accepting a new cmd connection")
                    (newSock, clientAddr) = readSock.accept()
                    newSock.setblocking(0)
                    inputs.append(newSock)
                    inputBufs[newSock] = ""
                else:
                    # Read data from this socket
                    logger.debug("Reading data from the command socket")
                    try:
                        inputBuf = readSock.recv(2048)
                    except:
                        inputBuf = 0
                    if inputBuf:
                        # Add the data read to this socket's receive buffer
                        inputBufs[readSock] += inputBuf
                        while chr(0) in inputBufs[readSock]:
                            inEnd = inputBufs[readSock].find(chr(0))
                            inCmd = inputBufs[readSock][:inEnd]
                            inputBufs[readSock] = inputBufs[readSock][inEnd+1:]
                            logger.debug("    %s" % (inCmd,))
                            self.HandleCommand(inCmd, readSock)
                    else:
                        # If no data was available, the socket has been closed, end of message.
                        if readSock in inputs:
                            logger.debug("Closing a socket because there's nothing to read")
                            del inputBufs[readSock]
                            inputs.remove(readSock)
                            readSock.close()

            for writeSock in writable:
                self.cmdOutputLock.acquire()
                writeStr = self.cmdOutputBufs.get(writeSock)
                self.cmdOutputLock.release()
                logger.debug("Sending data to a socket: %s" % (writeStr,))
                if writeStr is not None:
                    try:
                        numSent = writeSock.send(writeStr)
                    except:
                        logger.debug("Error sending data to the socket. Closing socket.")
                        numSent = len(writeStr)
                        writeSock.close()
                    self.cmdOutputLock.acquire()
                    self.cmdOutputBufs[writeSock] = writeStr[numSent:]
                    if not self.cmdOutputBufs[writeSock]:
                        logger.debug("No more data to output right now on the socket")
                        del self.cmdOutputBufs[writeSock]
                    self.cmdOutputLock.release()

            for errSock in exceptional:
                logger.debug("Closing a socket because of an error")
                if errSock in inputs:
                    if errSock in inputBufs:
                        del inputBufs[errSock]
                    inputs.remove(errSock)
                self.cmdOutputLock.acquire()
                if errSock in self.cmdOutputBufs:
                    del self.cmdOutputBufs[errSock]
                self.cmdOutputLock.release()
                errSock.close()
                if errSock is self.cmdSrvSock:
                    logger.debug("   The socket was the server socket")
                    self.cmdSrvSock = None

            # Check to see if any sockets have closed
            self.cmdOutputLock.acquire()
            allSocks = self.cmdOutputBufs.keys() + inputs
            self.cmdOutputLock.release()

            for testSock in allSocks:
                try:
                    testSock.fileno()
                except socket.error:
                    self.cmdOutputLock.acquire()
                    self.cmdOutputBufs.pop(testSock, None)
                    self.cmdOutputLock.release()
                    if testSock in inputs:
                        inputs.remove(testSock)

        logger.debug("Finished execution of the thread CmdRecv")

    def todict(self, obj, classkey=None):
        if isinstance(obj, dict):
            data = {}
            for (k, v) in obj.items():
                data[k] = self.todict(v, classkey)
            return data
        elif hasattr(obj, "_ast"):
            return self.todict(obj._ast())
        elif hasattr(obj, "__iter__"):
            return [self.todict(v, classkey) for v in obj]
        elif hasattr(obj, "__dict__"):
            data = dict([(key, self.todict(value, classkey)) 
                for key, value in obj.__dict__.iteritems() 
                if not callable(value) and not key.startswith('_')])
            if classkey is not None and hasattr(obj, "__class__"):
                data[classkey] = obj.__class__.__name__
            return data
        else:
            return obj

    def slotToFanCtrl(self, slotName):
        slotId = self.chassis.GetSlotNum()
        slotToFanCtrl = {}
        if slotId in [33, 35]:
            slotToFanCtrl = { "fc102" : (0, "fca"), "fc202" : (0, "fcb"), "lc501" : (0, "lca"),
                              "lc601" : (0, "lcb"), "lc701" : (0, "lcc"), "lc801" : (0, "lcd"),
                              "fc101" : (1, "fca"), "fc201" : (1, "fcb"), "lc101" : (1, "lca"),
                              "lc201" : (1, "lcb"), "lc301" : (1, "lcc"), "lc401" : (1, "lcd") }
        elif slotId in [37, 39]:
            slotToFanCtrl = { "fc302" : (0, "fca"), "fc402" : (0, "fcb"), "lc102" : (0, "lca"),
                              "lc202" : (0, "lcb"), "lc302" : (0, "lcc"), "lc402" : (0, "lcd"),
                              "fc301" : (1, "fca"), "fc401" : (1, "fcb"), "lc502" : (1, "lca"),
                              "lc602" : (1, "lcb"), "lc702" : (1, "lcc"), "lc802" : (1, "lcd") }
        elif slotId in [41, 43]:
            slotToFanCtrl = { "fc502"  : (0, "fca"), "fc602"  : (0, "fcb"), "lc1301" : (0, "lca"),
                              "lc1401" : (0, "lcb"), "lc1501" : (0, "lcc"), "lc1601" : (0, "lcd"),
                              "fc501"  : (1, "fca"), "fc601"  : (1, "fcb"), "lc901"  : (1, "lca"),
                              "lc1001" : (1, "lcb"), "lc1101" : (1, "lcc"), "lc1201" : (1, "lcd") }
        elif slotId in [45, 47]:
            slotToFanCtrl = { "fc702"  : (0, "fca"), "fc802"  : (0, "fcb"), "lc902"  : (0, "lca"),
                              "lc1002" : (0, "lcb"), "lc1102" : (0, "lcc"), "lc1202" : (0, "lcd"),
                              "fc701"  : (1, "fca"), "fc801"  : (1, "fcb"), "lc1302" : (1, "lca"),
                              "lc1402" : (1, "lcb"), "lc1502" : (1, "lcc"), "lc1602" : (1, "lcd") }
        return slotToFanCtrl.get(slotName, (0, ""))

    #-------------------------------------------------------------------------------
    #
    #   Command Interface Functions - These functions implement the command interface
    #   to the chassisd daemon.
    #
    #-------------------------------------------------------------------------------

    def HandleCmdEcho(self, cmdStr, ourSock):
        return { 'reply' : cmdStr[5:] }

    def HandleCmdSummary(self, cmdStr, ourSock):
        if self.actchd is None:
            retVal = { 'noCtp' : 'notRunning' }
        elif self.actchd.ctpData is None:
            if self.actchd.cmdargs.chasStaticTopo:
                retVal = { 'noCtp' : 'staticTopo' }
            else:
                retVal = { 'noCtp' : 'cpuB' }
        else:
            retVal = self.todict(self.actchd.ctpData)
        return retVal

    def HandleSetLogLevel(self, cmdStr, ourSock):
        retVal = {}
        cmdList = cmdStr.split()
        try:
            logLevel = getattr(logging, cmdList[1].upper(), None)
            if not isinstance(logLevel, int):
                retVal = { 'errorMsg' : "Invalid log level: %s" % (cmdList[1],) }
            logger.setLevel(logLevel)
        except (TypeError, IndexError):
            retVal = { 'errorMsg' : "Invalid arguments: %s" % (cmdStr,) }
        return retVal

    def HandleGetTemps(self, cmdStr, ourSock):
        if self.actchd is None:
            retVal = { 'noTemps' : 'notRunning' }
        elif self.actchd.tempData is None:
            retVal = { 'noTemps' : 'notLM' }
        else:
            retVal = self.todict(self.actchd.tempData)
        return retVal

    # Set a temperature threshold: '<card> <sensor> [warn|shut] <value>'
    def HandleSetTempThresh(self, cmdStr, ourSock):
        retVal = {}
        cmdList = cmdStr.split()
        if self.actchd is None:
            retVal = { 'errorMsg' : 'accton-chassisd is not running' }
        elif self.actchd.tempData is None:
            retVal = { 'errorMsg' : 'must be run on local master fabric card' }
        elif len(cmdList) != 5:
            retVal = { 'errorMsg' : "Incorrect number of parameters" }
        else:
            fctrl, card = self.slotToFanCtrl(cmdList[1].lower())
            if card == "":
                retVal = { 'errorMsg' : "The card name %s is invalid" % (cmdList[1],) }
            else:
                fanCtrl = self.actchd.tempData.lcd[fctrl]
                if card == "fca":
                    fanCtrl = self.actchd.tempData.fca[fctrl]
                elif card == "fcb":
                    fanCtrl = self.actchd.tempData.fcb[fctrl]
                elif card == "lca":
                    fanCtrl = self.actchd.tempData.lca[fctrl]
                elif card == "lcb":
                    fanCtrl = self.actchd.tempData.lcb[fctrl]
                elif card == "lcc":
                    fanCtrl = self.actchd.tempData.lcc[fctrl]

                td = None
                if cmdList[2].lower() == "asic":
                    td = fanCtrl.asic
                elif cmdList[2].lower() == "cpu":
                    td = fanCtrl.cpu
                elif cmdList[2].lower() == "lm75a":
                    td = fanCtrl.lm75a
                elif cmdList[2].lower() == "lm75b":
                    td = fanCtrl.lm75b
                elif cmdList[2].lower() == "lm75c":
                    td = fanCtrl.lm75c
                elif cmdList[2].lower() == "lm75d":
                    td = fanCtrl.lm75d
                if td is None:
                    retVal = { 'errorMsg' : "The sensor name %s is invalid" % (cmdList[2],) }
                else:
                    try:
                        thresh = int(cmdList[4], 0)
                    except ValueError, TypeError:
                        retVal = { 'errorMsg' : "The threshold value %s is invalid" % (cmdList[4],) }
                    else:
                        if cmdList[3].lower() == "warn":
                            td.warn = thresh
                        elif cmdList[3].lower() == "shut":
                            td.shutdown = thresh
                        else:
                            retVal = { 'errorMsg' : "The threshold must be 'warn' or 'shut', not %s" % (cmdList[3],) }
        return retVal

    def HandleCmdLocalMasters(self, cmdStr, ourSock):
        if self.actchd is None:
            retVal = { 'noLM' : 'notRunning' }
        elif self.actchd.localMasters == [0] * 4:
            retVal = { 'noLM' : 'cpuB' }
        else:
            retVal = self.todict(self.actchd.localMasters)
        return retVal

    def SendSock(self, outSock, outStr):
        logger.debug("Sending data to a socket: %s" % (outStr,))
        if outStr is not None:
            if not self.cmdOutputBufs.get(outSock):
                try:
                    numSent = outSock.send(outStr)
                except:
                    logger.debug("Error sending data to the socket. Closing socket.")
                    numSent = len(outStr)
                    outSock.close()
                outStr = outStr[numSent:]
            if outStr:
                self.cmdOutputLock.acquire()
                self.cmdOutputBufs.setdefault(outSock, "")
                self.cmdOutputBufs[outSock] += outStr
                self.cmdOutputLock.release()
        return

    def HandleCommand(self, cmdStr, ourSock):
        '''
        This function is called from the CmdRecv thread when a command has been
        received. The command is parsed, executed,  and the output data, if any, 
        is added to the output buffer.
        '''

        chassisdCmds = {
        #   Command              Function                    Priviledged
            "echo"            : (self.HandleCmdEcho,         False),
            "getsummary"      : (self.HandleCmdSummary,      False),
            "setloglevel"     : (self.HandleSetLogLevel,     True),
            "gettemps"        : (self.HandleGetTemps,        False),
            "settempthresh"   : (self.HandleSetTempThresh,   True),
            "localmasters"    : (self.HandleCmdLocalMasters, False),
        }

        ucredStr = ourSock.getsockopt(socket.SOL_SOCKET, chassisd_cmdsrv._SOCKET_PEERCRED, struct.calcsize("3I"))
        (peerPid, peerUid, peerGid) = struct.unpack("3I", ucredStr)
        canModify = (peerUid == 0)

        cmdList = cmdStr.split()
        if cmdList[0].lower() in chassisdCmds:
            if chassisdCmds[cmdList[0].lower()][1] and not canModify:
                outDict = { 'errorMsg' : "ERROR You do not have sufficient privilege to execute this command: %s" % (cmdStr,) }
            else:
                outDict = chassisdCmds[cmdList[0].lower()][0](cmdStr, ourSock)
        else:
             outDict = { 'errorMsg' : "ERROR Unknown command: %s" % (cmdStr,) }

        jsonStr = json.dumps(outDict, sort_keys=not self.cmdargs.cmdNoSort, indent=self.cmdargs.cmdIndent)
        self.SendSock(ourSock, jsonStr + chr(0))


if __name__ == '__main__':
    pass
